import math
from abjad import mathtools
from . import constants
from abjad.system.FormatSpecification import FormatSpecification
from .Pitch import Pitch


class NamedPitch(Pitch):
    r"""
    Named pitch.

    ..  container:: example

        Initializes from pitch name:

        >>> pitch = abjad.NamedPitch("cs''")
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "treble"
                cs''1 * 1/4
            }

        Initializes quartertone from pitch name:

        >>> pitch = abjad.NamedPitch("aqs")
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "bass"
                aqs1 * 1/4
            }

    ..  container:: example

        Initializes from pitch-class / octave string:

        >>> pitch = abjad.NamedPitch('C#5')
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "treble"
                cs''1 * 1/4
            }

        Initializes quartertone from pitch-class / octave string:

        >>> pitch = abjad.NamedPitch('A+3')
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "bass"
                aqs1 * 1/4
            }

        >>> pitch = abjad.NamedPitch('Aqs3')
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "bass"
                aqs1 * 1/4
            }

    ..  container:: example

        Initializes arrowed pitch:

        >>> pitch = abjad.NamedPitch('C#5', arrow=abjad.Up)
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \once \override Accidental.stencil = #ly:text-interface::print
                \once \override Accidental.text = \markup { \musicglyph #"accidentals.sharp.arrowup" }
                \clef "treble"
                cs''1 * 1/4
            }

    ..  container:: example

        REGRESSION. Small floats just less than a C initialize in the correct
        octave.

        Initializes c / C3:

        >>> pitch = abjad.NamedPitch(-12.1)
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "bass"
                c1 * 1/4
            }

        Initializes c' / C4:

        >>> pitch = abjad.NamedPitch(-0.1)
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "treble"
                c'1 * 1/4
            }

        Initializes c'' / C5:

        >>> pitch = abjad.NamedPitch(11.9)
        >>> abjad.show(pitch) # doctest: +SKIP

        ..  docs::

            >>> staff = pitch.__illustrate__()[abjad.Staff]
            >>> abjad.f(staff)
            \new Staff
            \with
            {
                \override TimeSignature.stencil = ##f
            }
            {
                \clef "treble"
                c''1 * 1/4
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(
        self,
        name="c'",
        *,
        accidental=None,
        arrow=None,
        octave=None,
    ):
        super().__init__(
            name or "c'",
            accidental=accidental,
            arrow=arrow,
            octave=octave,
            )

    ### SPECIAL METHODS ###

    def __add__(self, interval):
        """
        Adds named pitch to `interval`.

        ..  container:: example

            >>> abjad.NamedPitch("cs''") + abjad.NamedInterval('-M2')
            NamedPitch("b'")

            >>> abjad.NamedPitch("cs''") + abjad.NamedInterval('P1')
            NamedPitch("cs''")

            >>> abjad.NamedPitch("cs''") + abjad.NamedInterval('+M2')
            NamedPitch("ds''")

        Returns new named pitch.
        """
        import abjad
        interval = abjad.NamedInterval(interval)
        return interval.transpose(self)

    def __copy__(self, *arguments):
        """
        Copies named pitch.

        >>> import copy

        ..  container:: example

            >>> copy.copy(abjad.NamedPitch("c''"))
            NamedPitch("c''")

            >>> copy.copy(abjad.NamedPitch("cs''"))
            NamedPitch("cs''")

            >>> copy.copy(abjad.NamedPitch("df''"))
            NamedPitch("df''")

        ..  container:: example

            Copies arrowed pitch:

            >>> pitch = abjad.NamedPitch("cs''", arrow=abjad.Up)
            >>> copy.copy(pitch)
            NamedPitch("cs''", arrow=Up)

        Returns new named pitch.
        """
        return type(self)(self, arrow=self.arrow)

    def __eq__(self, argument):
        """
        Is true when `argument` is a named pitch equal to this named pitch.

        ..  container:: example

            >>> pitch_1 = abjad.NamedPitch('fs')
            >>> pitch_2 = abjad.NamedPitch('fs')
            >>> pitch_3 = abjad.NamedPitch('gf')

            >>> pitch_1 == pitch_1
            True
            >>> pitch_1 == pitch_2
            True
            >>> pitch_1 == pitch_3
            False

            >>> pitch_2 == pitch_1
            True
            >>> pitch_2 == pitch_2
            True
            >>> pitch_2 == pitch_3
            False

            >>> pitch_3 == pitch_1
            False
            >>> pitch_3 == pitch_2
            False
            >>> pitch_3 == pitch_3
            True

        Returns true or false.
        """
        return super().__eq__(argument)

    def __hash__(self):
        """
        Hashes named pitch.

        Returns integer.
        """
        return super().__hash__()

    def __lt__(self, argument):
        """
        Is true when named pitch is less than `argument`.

        ..  container:: example

            >>> pitch_1 = abjad.NamedPitch('fs')
            >>> pitch_2 = abjad.NamedPitch('fs')
            >>> pitch_3 = abjad.NamedPitch('gf')

            >>> pitch_1 < pitch_1
            False
            >>> pitch_1 < pitch_2
            False
            >>> pitch_1 < pitch_3
            True

            >>> pitch_2 < pitch_1
            False
            >>> pitch_2 < pitch_2
            False
            >>> pitch_2 < pitch_3
            True

            >>> pitch_3 < pitch_1
            False
            >>> pitch_3 < pitch_2
            False
            >>> pitch_3 < pitch_3
            False

        Returns true or false.
        """
        try:
            argument = type(self)(argument)
        except (TypeError, ValueError):
            return False
        self_dpn = self._get_diatonic_pitch_number()
        argument_dpn = argument._get_diatonic_pitch_number()
        if self_dpn == argument_dpn:
            return self.accidental < argument.accidental
        return self_dpn < argument_dpn

    def __radd__(self, interval):
        """
        Right-addition not defined on named pitches.

        ..  container:: example

            >>> abjad.NamedPitch("cs'").__radd__(1)
            Traceback (most recent call last):
                ...
            NotImplementedError: right-addition not defined on NamedPitch.

        """
        message = 'right-addition not defined on {}.'
        message = message.format(type(self).__name__)
        raise NotImplementedError(message)

    def __str__(self):
        """
        Gets string representation of named pitch.

        ..  container:: example

            >>> str(abjad.NamedPitch("c''"))
            "c''"

            >>> str(abjad.NamedPitch("cs''"))
            "cs''"

            >>> str(abjad.NamedPitch("df''"))
            "df''"

        Returns string.
        """
        return self.name

    def __sub__(self, argument):
        """
        Subtracts `argument` from named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("cs''") - abjad.NamedPitch("b'")
            NamedInterval('-M2')

            >>> abjad.NamedPitch("cs''") - abjad.NamedPitch("fs''")
            NamedInterval('+P4')

        Returns named interval.
        """
        import abjad
        if isinstance(argument, type(self)):
            return abjad.NamedInterval.from_pitch_carriers(self, argument)
        interval = abjad.NamedInterval(argument)
        interval = -interval
        return interval.transpose(self)

    ### PRIVATE METHODS ###

    def _apply_accidental(self, accidental):
        import abjad
        name = self._get_diatonic_pc_name()
        name += str(self.accidental + abjad.Accidental(accidental))
        name += self.octave.ticks
        return type(self)(name)

    def _from_named_parts(self, dpc_number, alteration, octave):
        import abjad
        dpc_name = constants._diatonic_pc_number_to_diatonic_pc_name[dpc_number]
        accidental = abjad.Accidental(alteration)
        octave = abjad.Octave(octave)
        self._octave = octave
        self._pitch_class = abjad.NamedPitchClass(dpc_name + str(accidental))

    def _from_number(self, number):
        import abjad
        number = self._to_nearest_quarter_tone(number)
        div, mod = divmod(number, 12)
        pitch_class = abjad.NumberedPitchClass(mod)
        self._from_named_parts(
            dpc_number=pitch_class._get_diatonic_pc_number(),
            alteration=pitch_class._get_alteration(),
            octave=div + 4,
            )

    def _from_pitch_or_pitch_class(self, pitch_or_pitch_class):
        import abjad
        name = format(pitch_or_pitch_class, 'lilypond')
        if not isinstance(pitch_or_pitch_class, Pitch):
            name += "'"
        if isinstance(pitch_or_pitch_class, Pitch):
            self._pitch_class = abjad.NamedPitchClass(
                pitch_or_pitch_class.pitch_class,
                )
            self._octave = pitch_or_pitch_class.octave
        else:
            self._pitch_class = abjad.NamedPitchClass(
                pitch_or_pitch_class,
                )
            self._octave = abjad.Octave()

    def _get_alteration(self):
        return self.accidental.semitones

    def _get_diatonic_pc_name(self):
        return constants._diatonic_pc_number_to_diatonic_pc_name[
            self.pitch_class._diatonic_pc_number]

    def _get_diatonic_pc_number(self):
        diatonic_pc_name = self._get_diatonic_pc_name()
        diatonic_pc_number = (
            constants._diatonic_pc_name_to_diatonic_pc_number[
                diatonic_pc_name]
            )
        return diatonic_pc_number

    def _get_diatonic_pitch_number(self):
        diatonic_pitch_number = 7 * (self.octave.number - 4)
        diatonic_pitch_number += self._get_diatonic_pc_number()
        return diatonic_pitch_number

    def _get_format_specification(self):
        return FormatSpecification(
            self,
            coerce_for_equality=True,
            repr_is_indented=False,
            storage_format_args_values=[self.name],
            storage_format_is_indented=False,
            storage_format_kwargs_names=['arrow'],
            )

    def _get_lilypond_format(self):
        return str(self)

    def _list_format_contributions(self):
        contributions = []
        if self.arrow is None:
            return contributions
        string = r'\once \override Accidental.stencil ='
        string += ' #ly:text-interface::print'
        contributions.append(string)
        glyph = f'accidentals.{self.accidental.name}'
        glyph += f'.arrow{str(self.arrow).lower()}'
        string = r'\once \override Accidental.text ='
        string += rf' \markup {{ \musicglyph #"{glyph}" }}'
        contributions.append(string)
        return contributions

    def _respell_with_flats(self):
        name = constants._pitch_class_number_to_pitch_class_name_with_flats[
            self.pitch_class.number
            ]
        pitch = type(self)((name, self.octave.number))
        return pitch

    def _respell_with_sharps(self):
        name = constants._pitch_class_number_to_pitch_class_name_with_sharps[
            self.pitch_class.number
            ]
        pitch = type(self)((name, self.octave.number))
        return pitch

    ### PUBLIC PROPERTIES ###

    @property
    def accidental(self):
        """
        Gets accidental of named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("c''").accidental
            Accidental('natural')

            >>> abjad.NamedPitch("cs''").accidental
            Accidental('sharp')

            >>> abjad.NamedPitch("df''").accidental
            Accidental('flat')

        Returns accidental.
        """
        return self.pitch_class.accidental

    @property
    def arrow(self):
        """
        Gets arrow of named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("cs''").arrow is None
            True

            >>> abjad.NamedPitch("cs''", arrow=abjad.Up).arrow
            Up

            >>> abjad.NamedPitch("cs''", arrow=abjad.Down).arrow
            Down

        ..  container:: example

            Displays arrow in interpreter representation:

            >>> abjad.NamedPitch("cs''", arrow=abjad.Down)
            NamedPitch("cs''", arrow=Down)

        Returns up, down or none.
        """
        return self._pitch_class.arrow

    @property
    def hertz(self):
        """
        Gets frequency of named pitch in Hertz.

        ..  container:: example

            >>> abjad.NamedPitch("c''").hertz
            523.25...

            >>> abjad.NamedPitch("cs''").hertz
            554.36...

            >>> abjad.NamedPitch("df''").hertz
            554.36...

        Returns float.
        """
        return super().hertz

    @property
    def name(self):
        """
        Gets name of named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("c''").name
            "c''"

            >>> abjad.NamedPitch("cs''").name
            "cs''"

            >>> abjad.NamedPitch("df''").name
            "df''"

        Returns string.
        """
        return '{!s}{!s}'.format(
            self.pitch_class,
            self.octave,
            )

    @property
    def number(self):
        """
        Gets number of named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("c''").number
            12

            >>> abjad.NamedPitch("cs''").number
            13

            >>> abjad.NamedPitch("df''").number
            13

            >>> abjad.NamedPitch("cf'").number
            -1

        Returns number.
        """
        diatonic_pc_number = self.pitch_class._get_diatonic_pc_number()
        pc_number = constants._diatonic_pc_number_to_pitch_class_number[diatonic_pc_number]
        alteration = self.pitch_class._get_alteration()
        octave_base_pitch = (self.octave.number - 4) * 12
        return mathtools.integer_equivalent_number_to_integer(
            pc_number + alteration + octave_base_pitch
            )

    @property
    def octave(self):
        """
        Gets octave of named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("c''").octave
            Octave(5)

            >>> abjad.NamedPitch("cs''").octave
            Octave(5)

            >>> abjad.NamedPitch("df''").octave
            Octave(5)

        Returns octave.
        """
        return self._octave

    @property
    def pitch_class(self):
        """
        Gets pitch-class of named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("c''").pitch_class
            NamedPitchClass('c')

            >>> abjad.NamedPitch("cs''").pitch_class
            NamedPitchClass('cs')

            >>> abjad.NamedPitch("df''").pitch_class
            NamedPitchClass('df')

        Returns named pitch-class.
        """
        return self._pitch_class

    ### PUBLIC METHODS ###

    @classmethod
    def from_hertz(class_, hertz):
        """
        Makes named pitch from `hertz`.

        ..  container:: example

            >>> abjad.NamedPitch.from_hertz(440)
            NamedPitch("a'")

        ..  container:: example

            REGRESSION. Returns c'' (C5) and not c' (C4):

            >>> abjad.NamedPitch.from_hertz(519)
            NamedPitch("c''")

        Returns newly constructed named pitch.
        """
        return super().from_hertz(hertz)

    def get_name(self, locale=None):
        """
        Gets name of named pitch according to `locale`.

        ..  container:: example

            >>> abjad.NamedPitch("cs''").get_name()
            "cs''"

            >>> abjad.NamedPitch("cs''").get_name(locale='us')
            'C#5'

        Set `locale` to `'us'` or none.

        Returns string.
        """
        if locale is None:
            return self.name
        elif locale == 'us':
            return '{}{}{}'.format(
                self._get_diatonic_pc_name().upper(),
                self.accidental.symbol,
                self.octave.number,
                )
        else:
            message = "must be 'us' or none: {!r}."
            message = message.format(locale)
            raise ValueError(message)

    def invert(self, axis=None):
        """
        Inverts named pitch around `axis`.

        ..  container:: example

            Inverts pitch around middle C explicitly:

            >>> abjad.NamedPitch("d'").invert("c'")
            NamedPitch('bf')

            >>> abjad.NamedPitch('bf').invert("c'")
            NamedPitch("d'")

        ..  container:: example

            Inverts pitch around middle C implicitly:

            >>> abjad.NamedPitch("d'").invert()
            NamedPitch('bf')

            >>> abjad.NamedPitch('bf').invert()
            NamedPitch("d'")

        ..  container:: example

            Inverts pitch around A3:

            >>> abjad.NamedPitch("d'").invert('a')
            NamedPitch('e')

        Interprets none-valued `axis` equal to middle C.

        Returns new named pitch.
        """
        return super().invert(axis=axis)

    def multiply(self, n=1):
        """
        Multiplies named pitch.

        ..  container:: example

            >>> abjad.NamedPitch("d'").multiply(1)
            NamedPitch("d'")

            >>> abjad.NamedPitch("d'").multiply(3)
            NamedPitch("fs'")

            >>> abjad.NamedPitch("d'").multiply(6)
            NamedPitch("c''")

            >>> abjad.NamedPitch("d'").multiply(6.5)
            NamedPitch("cs''")

        Returns new named pitch.
        """
        return super().multiply(n=n)

    def simplify(self):
        """
        Reduce alteration to between -2 and 2 while maintaining identical pitch
        number.

            >>> abjad.NamedPitch("cssqs'").simplify()
            NamedPitch("dqs'")

            >>> abjad.NamedPitch("cfffqf'").simplify()
            NamedPitch('aqf')

            >>> float(abjad.NamedPitch("cfffqf'").simplify()) == float(NamedPitch('aqf'))
            True

        ..  note:: LilyPond by default only supports accidentals from
                   double-flat to double-sharp.

        Returns named pitch.
        """
        import abjad
        alteration = self._get_alteration()
        if abs(alteration) <= 2:
            return self
        diatonic_pc_number = self._get_diatonic_pc_number()
        octave = int(self.octave)
        while alteration > 2:
            step_size = 2
            if diatonic_pc_number == 2:  # e to f
                step_size = 1
            elif diatonic_pc_number == 6:  # b to c
                step_size = 1
                octave += 1
            diatonic_pc_number = (diatonic_pc_number + 1) % 7
            alteration -= step_size
        while alteration < -2:
            step_size = 2
            if diatonic_pc_number == 3:  # f to e
                step_size = 1
            elif diatonic_pc_number == 0:  # c to b
                step_size = 1
                octave -= 1
            diatonic_pc_number = (diatonic_pc_number - 1) % 7
            alteration += step_size
        diatonic_pc_name = constants._diatonic_pc_number_to_diatonic_pc_name[
            diatonic_pc_number]
        accidental = abjad.Accidental(alteration)
        octave = abjad.Octave(octave)
        pitch_name = '{}{!s}{!s}'.format(diatonic_pc_name, accidental, octave)
        return type(self)(pitch_name, arrow=self.arrow)

    # TODO: duplicate on NumberedPitch
    def to_staff_position(self, clef=None):
        r"""
        Changes named pitch to staff position.

        ..  container:: example

            Changes C#5 to absolute staff position:

            >>> abjad.NamedPitch('C#5').to_staff_position()
            StaffPosition(7)

            >>> abjad.NamedPitch('C#5').to_staff_position(clef='treble')
            StaffPosition(1)

            >>> abjad.NamedPitch('C#5').to_staff_position(clef='bass')
            StaffPosition(13)

        ..  container:: example

            Marks up absolute staff position of many pitches:

            >>> staff = abjad.Staff("g16 a b c' d' e' f' g' a' b' c'' d'' e'' f'' g'' a''")
            >>> for note in staff:
            ...     staff_position = note.written_pitch.to_staff_position()
            ...     markup = abjad.Markup(staff_position.number)
            ...     abjad.attach(markup, note)
            ...
            >>> abjad.override(staff).text_script.staff_padding = 5
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = #5
                }
                {
                    g16
                    - \markup { -3 }
                    a16
                    - \markup { -2 }
                    b16
                    - \markup { -1 }
                    c'16
                    - \markup { 0 }
                    d'16
                    - \markup { 1 }
                    e'16
                    - \markup { 2 }
                    f'16
                    - \markup { 3 }
                    g'16
                    - \markup { 4 }
                    a'16
                    - \markup { 5 }
                    b'16
                    - \markup { 6 }
                    c''16
                    - \markup { 7 }
                    d''16
                    - \markup { 8 }
                    e''16
                    - \markup { 9 }
                    f''16
                    - \markup { 10 }
                    g''16
                    - \markup { 11 }
                    a''16
                    - \markup { 12 }
                }

        ..  container:: example

            Marks up treble staff position of many pitches:

            >>> staff = abjad.Staff("g16 a b c' d' e' f' g' a' b' c'' d'' e'' f'' g'' a''")
            >>> clef = abjad.Clef('treble')
            >>> for note in staff:
            ...     staff_position = note.written_pitch.to_staff_position(
            ...         clef=clef
            ...         )
            ...     markup = abjad.Markup(staff_position.number)
            ...     abjad.attach(markup, note)
            ...
            >>> abjad.override(staff).text_script.staff_padding = 5
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = #5
                }
                {
                    g16
                    - \markup { -9 }
                    a16
                    - \markup { -8 }
                    b16
                    - \markup { -7 }
                    c'16
                    - \markup { -6 }
                    d'16
                    - \markup { -5 }
                    e'16
                    - \markup { -4 }
                    f'16
                    - \markup { -3 }
                    g'16
                    - \markup { -2 }
                    a'16
                    - \markup { -1 }
                    b'16
                    - \markup { 0 }
                    c''16
                    - \markup { 1 }
                    d''16
                    - \markup { 2 }
                    e''16
                    - \markup { 3 }
                    f''16
                    - \markup { 4 }
                    g''16
                    - \markup { 5 }
                    a''16
                    - \markup { 6 }
                }

        ..  container:: example

            Marks up bass staff position of many pitches:

            >>> staff = abjad.Staff("g,16 a, b, c d e f g a b c' d' e' f' g' a'")
            >>> clef = abjad.Clef('bass')
            >>> abjad.attach(clef, staff[0])
            >>> for note in staff:
            ...     staff_position = note.written_pitch.to_staff_position(
            ...         clef=clef
            ...         )
            ...     markup = abjad.Markup(staff_position.number)
            ...     abjad.attach(markup, note)
            ...
            >>> abjad.override(staff).text_script.staff_padding = 5
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = #5
                }
                {
                    \clef "bass"
                    g,16
                    - \markup { -4 }
                    a,16
                    - \markup { -3 }
                    b,16
                    - \markup { -2 }
                    c16
                    - \markup { -1 }
                    d16
                    - \markup { 0 }
                    e16
                    - \markup { 1 }
                    f16
                    - \markup { 2 }
                    g16
                    - \markup { 3 }
                    a16
                    - \markup { 4 }
                    b16
                    - \markup { 5 }
                    c'16
                    - \markup { 6 }
                    d'16
                    - \markup { 7 }
                    e'16
                    - \markup { 8 }
                    f'16
                    - \markup { 9 }
                    g'16
                    - \markup { 10 }
                    a'16
                    - \markup { 11 }
                }

        Returns staff position.
        """
        import abjad
        staff_position_number = self._get_diatonic_pitch_number()
        if clef is not None:
            clef = abjad.Clef(clef)
            staff_position_number += clef.middle_c_position.number
        staff_position = abjad.StaffPosition(staff_position_number)
        return staff_position

    def transpose(self, n=0):
        """
        Transposes named pitch by index `n`.

        ..  container:: example

            Transposes C4 up a minor second:

            >>> abjad.NamedPitch("c'").transpose(n='m2')
            NamedPitch("df'")

        ..  container:: example

            Transposes C4 down a major second:

            >>> abjad.NamedPitch("c'").transpose(n='-M2')
            NamedPitch('bf')

        Returns new named pitch.
        """
        import abjad
        interval = abjad.NamedInterval(n)
        pitch_number = self.number + interval.semitones
        diatonic_pc_number = self._get_diatonic_pc_number()
        diatonic_pc_number += interval.staff_spaces
        diatonic_pc_number %= 7
        diatonic_pc_name = \
            constants._diatonic_pc_number_to_diatonic_pc_name[
                diatonic_pc_number]
        pc = constants._diatonic_pc_name_to_pitch_class_number[
            diatonic_pc_name
            ]
        nearest_neighbor = self._to_nearest_octave(pitch_number, pc)
        semitones = pitch_number - nearest_neighbor
        accidental = abjad.Accidental(semitones)
        octave = int(math.floor((pitch_number - semitones) / 12)) + 4
        octave = abjad.Octave(octave)
        name = diatonic_pc_name + str(accidental) + octave.ticks
        return type(self)(name)



import copy
import os
import subprocess
import uqbar.graphs


def graph(
    argument,
    image_format='pdf',
    layout='dot',
    graph_attributes=None,
    node_attributes=None,
    edge_attributes=None,
    **keywords
    ) -> None:
    r"""
    Graphs ``argument``.

    ..  container:: example

        Graphs staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.graph(staff) # doctest: +SKIP

        ..  docs::

            >>> print(format(staff.__graph__(), 'graphviz'))
            digraph G {
                graph [style=rounded];
                node [fontname=Arial,
                    shape=none];
                Staff_0 [label=<
                    <TABLE BORDER="2" CELLPADDING="5">
                        <TR>
                            <TD BORDER="0">Staff</TD>
                        </TR>
                    </TABLE>>,
                    margin=0.05,
                    style=rounded];
                subgraph Staff {
                    graph [color=grey75,
                        penwidth=2];
                    Note_0 [label=<
                        <TABLE BORDER="2" CELLPADDING="5">
                            <TR>
                                <TD BORDER="0">Note</TD>
                            </TR>
                            <HR/>
                            <TR>
                                <TD BORDER="0">c'4</TD>
                            </TR>
                        </TABLE>>,
                        margin=0.05,
                        style=rounded];
                    Note_1 [label=<
                        <TABLE BORDER="2" CELLPADDING="5">
                            <TR>
                                <TD BORDER="0">Note</TD>
                            </TR>
                            <HR/>
                            <TR>
                                <TD BORDER="0">d'4</TD>
                            </TR>
                        </TABLE>>,
                        margin=0.05,
                        style=rounded];
                    Note_2 [label=<
                        <TABLE BORDER="2" CELLPADDING="5">
                            <TR>
                                <TD BORDER="0">Note</TD>
                            </TR>
                            <HR/>
                            <TR>
                                <TD BORDER="0">e'4</TD>
                            </TR>
                        </TABLE>>,
                        margin=0.05,
                        style=rounded];
                    Note_3 [label=<
                        <TABLE BORDER="2" CELLPADDING="5">
                            <TR>
                                <TD BORDER="0">Note</TD>
                            </TR>
                            <HR/>
                            <TR>
                                <TD BORDER="0">f'4</TD>
                            </TR>
                        </TABLE>>,
                        margin=0.05,
                        style=rounded];
                }
                Staff_0 -> Note_0;
                Staff_0 -> Note_1;
                Staff_0 -> Note_2;
                Staff_0 -> Note_3;
            }

    ..  container:: example

        Graphs rhythm tree:

        >>> rtm_syntax = '(3 ((2 (2 1)) 2))'
        >>> parser = abjad.rhythmtrees.RhythmTreeParser()
        >>> rhythm_tree = parser(rtm_syntax)[0]
        >>> abjad.graph(rhythm_tree) # doctest: +SKIP

        ..  docs::

            >>> print(format(rhythm_tree.__graph__(), 'graphviz'))
            digraph G {
                graph [bgcolor=transparent,
                    truecolor=true];
                node_0 [label="3",
                    shape=triangle];
                node_1 [label="2",
                    shape=triangle];
                node_2 [label="2",
                    shape=box];
                node_3 [label="1",
                    shape=box];
                node_4 [label="2",
                    shape=box];
                node_0 -> node_1;
                node_0 -> node_4;
                node_1 -> node_2;
                node_1 -> node_3;
            }

    Opens image in default image viewer.
    """
    from abjad import system
    from abjad import abjad_configuration

    if isinstance(argument, str):
        graphviz_format = argument
    else:
        if hasattr(argument, '__graph__'):
            graphviz_graph = argument.__graph__(**keywords)
        elif isinstance(argument, uqbar.graphs.Graph):
            graphviz_graph = copy.deepcopy(argument)
        else:
            raise TypeError('Cannot graph {!r}'.format(type(argument)))
        if graph_attributes:
            graphviz_graph.attributes.update(graph_attributes)
        if node_attributes:
            graphviz_graph.node_attributes.update(node_attributes)
        if edge_attributes:
            graphviz_graph.edge_attributes.update(edge_attributes)
        graphviz_format = format(graphviz_graph, 'graphviz')

    assert image_format in ('pdf', 'png')
    valid_layouts = (
        'circo',
        'dot',
        'fdp',
        'neato',
        'osage',
        'sfdp',
        'twopi',
        )
    assert layout in valid_layouts

    message = 'cannot find `{}` command-line tool.'
    message = message.format(layout)
    message += ' Please download Graphviz from graphviz.org.'
    assert system.IOManager.find_executable(layout), message

    ABJADOUTPUT = abjad_configuration['abjad_output_directory']
    system.IOManager._ensure_directory_existence(ABJADOUTPUT)
    dot_path = os.path.join(
        ABJADOUTPUT,
        system.IOManager.get_next_output_file_name(file_extension='dot'),
        )
    img_path = os.path.join(ABJADOUTPUT, dot_path.replace('dot', 'pdf'))

    with open(dot_path, 'w') as f:
        f.write(graphviz_format)

    command = '{} -v -T{} {} -o {}'
    command = command.format(layout, image_format, dot_path, img_path)
    subprocess.call(command, shell=True)

    pdf_viewer = abjad_configuration['pdf_viewer']
    ABJADOUTPUT = abjad_configuration['abjad_output_directory']
    system.IOManager.open_file(img_path, pdf_viewer)



import collections
from .TypedCollection import TypedCollection


class TypedFrozenset(TypedCollection, collections.Set):
    """
    Typed fozen set.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        TypedCollection.__init__(
            self,
            item_class=item_class,
            items=items,
            )
        items = items or []
        items = [self._item_coercer(_) for _ in items]
        self._collection = frozenset(items)

    ### SPECIAL METHODS ###

    def __and__(self, argument):
        """
        Logical AND of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__and__(argument._collection)
        result = type(self)(result)
        return result

    def __ge__(self, argument):
        """
        Is true when typed frozen set is greater than or equal to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__ge__(argument._collection)

    def __gt__(self, argument):
        """
        Is true when typed frozen set is greater than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__gt__(argument._collection)

    def __hash__(self):
        """
        Hashes typed frozen set.

        Returns integer.
        """
        import abjad
        hash_values = abjad.StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            message = 'unhashable type: {}'.format(self)
            raise TypeError(message)
        return result

    def __le__(self, argument):
        """
        Is true when typed frozen set is less than or equal to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__le__(argument._collection)

    def __lt__(self, argument):
        """
        Is true when typed frozen set is less than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__lt__(argument._collection)

    def __or__(self, argument):
        """
        Logical OR of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__or__(argument._collection)
        result = type(self)(result)
        return result

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from typed frozen set.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__sub__(argument._collection)
        result = type(self)(result)
        return result

    def __xor__(self, argument):
        """
        Logical XOR of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.__xor__(argument._collection)
        result = type(self)(result)
        return result

    ### PUBLIC METHODS ###

    def copy(self):
        """
        Copies typed frozen set.

        Returns new typed frozen set.
        """
        return type(self)(self._collection.copy())

    def difference(self, argument):
        """
        Typed frozen set set-minus ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.difference(argument._collection)
        result = type(self)(result)
        return result

    def intersection(self, argument):
        """
        Set-theoretic intersection of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.intersection(argument._collection)
        result = type(self)(result)
        return result

    def isdisjoint(self, argument):
        """
        Is true when typed frozen set shares no elements with ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.isdisjoint(argument._collection)

    def issubset(self, argument):
        """
        Is true when typed frozen set is a subset of ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.issubset(argument._collection)

    def issuperset(self, argument):
        """
        Is true when typed frozen set is a superset of ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.issuperset(argument._collection)

    def symmetric_difference(self, argument):
        """
        Symmetric difference of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.symmetric_difference(argument._collection)
        result = type(self)(result)
        return result

    def union(self, argument):
        """
        Union of typed frozen set and ``argument``.

        Returns new typed frozen set.
        """
        argument = type(self)(argument)
        result = self._collection.union(argument._collection)
        result = type(self)(result)
        return result



import functools
import itertools
from abjad.system import AbjadValueObject


class Enumerator(AbjadValueObject):
    """
    Enumerator.

    Collects combinatorial enumerator methods.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_sequence',
        )

    ### INITIALIZER ###

    def __init__(self, sequence=None):
        import abjad
        if sequence is not None:
            sequence = abjad.sequence(items=sequence)
        self._sequence = sequence

    ### PRIVATE METHODS ###

    @staticmethod
    def _is_restricted_growth_function(sequence):
        """
        Is true when ``sequence`` is a restricted growth function.

        ..  container:: example

            Is true when sequence is a restricted growth function:

            >>> abjad.Enumerator._is_restricted_growth_function([1, 1, 1, 1])
            True

            >>> abjad.Enumerator._is_restricted_growth_function([1, 1, 1, 2])
            True

            >>> abjad.Enumerator._is_restricted_growth_function([1, 1, 2, 1])
            True

            >>> abjad.Enumerator._is_restricted_growth_function([1, 1, 2, 2])
            True

        ..  container:: example

            Is false when sequence is not a restricted growth function:

            >>> abjad.Enumerator._is_restricted_growth_function([1, 1, 1, 3])
            False

            >>> abjad.Enumerator._is_restricted_growth_function([17])
            False

        A restricted growth function is a sequence ``l`` such that
        ``l[0] == 1`` and such that ``l[i] <= max(l[:i]) + 1`` for
        ``1 <= i <= len(l)``.

        Returns true or false.
        """
        try:
            for i, n in enumerate(sequence):
                if i == 0:
                    if not n == 1:
                        return False
                else:
                    if not n <= max(sequence[:i]) + 1:
                        return False
            return True
        except TypeError:
            return False

    @staticmethod
    def _partition_by_rgf(sequence, rgf):
        """
        Partitions ``sequence`` by restricted growth function ``rgf``.

        >>> sequence = abjad.sequence(range(10))
        >>> rgf = [1, 1, 2, 2, 1, 2, 3, 3, 2, 4]

        >>> abjad.Enumerator._partition_by_rgf(sequence, rgf)
        Sequence([Sequence([0, 1, 4]), Sequence([2, 3, 5, 8]), Sequence([6, 7]), Sequence([9])])

        Returns list of lists.
        """
        import abjad
        rgf = abjad.sequence(items=rgf)
        if not Enumerator._is_restricted_growth_function(rgf):
            message = 'must be restricted growth function: {!r}.'
            message = message.format(rgf)
            raise ValueError(message)
        if not len(sequence) == len(rgf):
            message = 'lengths must be equal.'
            raise ValueError(message)
        partition = []
        for part_index in range(max(rgf)):
            part = []
            partition.append(part)
        for n, part_number in zip(sequence, rgf):
            part_index = part_number - 1
            part = partition[part_index]
            part.append(n)
        partition = [abjad.sequence(_) for _ in partition]
        return abjad.sequence(items=partition)

    @staticmethod
    def _yield_restricted_growth_functions(length):
        """
        Yields restricted growth functions of ``length``.

        ..  container:: example

            >>> rgfs = abjad.Enumerator._yield_restricted_growth_functions(4)
            >>> for rgf in rgfs:
            ...     rgf
            ...
            (1, 1, 1, 1)
            (1, 1, 1, 2)
            (1, 1, 2, 1)
            (1, 1, 2, 2)
            (1, 1, 2, 3)
            (1, 2, 1, 1)
            (1, 2, 1, 2)
            (1, 2, 1, 3)
            (1, 2, 2, 1)
            (1, 2, 2, 2)
            (1, 2, 2, 3)
            (1, 2, 3, 1)
            (1, 2, 3, 2)
            (1, 2, 3, 3)
            (1, 2, 3, 4)

        Returns restricted growth functions in lex order.

        Returns generator of tuples.
        """
        import abjad
        assert abjad.mathtools.is_positive_integer(length), repr(length)
        last_rgf = list(range(1, length + 1))
        rgf = length * [1]
        yield tuple(rgf)
        while not rgf == last_rgf:
            for i, x in enumerate(reversed(rgf)):
                stop = -(i + 1)
                if x < max(rgf[:stop]) + 1:
                    first_part = rgf[:stop]
                    increased_part = [rgf[stop] + 1]
                    trailing_ones = i * [1]
                    rgf = first_part + increased_part + trailing_ones
                    yield tuple(rgf)
                    break

    ### PUBLIC PROPERTIES ###

    @property
    def sequence(self):
        """
        Gets sequence of enumerator.

        Defaults to none.

        Set to sequence or none.

        Returns sequence or none.
        """
        return self._sequence

    ### PUBLIC METHODS ###

    def yield_combinations(self, minimum_length=None, maximum_length=None):
        """
        Yields combinations of sequence items.

        ..  container:: example

            >>> enumerator = abjad.Enumerator([1, 2, 3, 4])
            >>> for combination in enumerator.yield_combinations():
            ...     combination
            Sequence([])
            Sequence([1])
            Sequence([2])
            Sequence([1, 2])
            Sequence([3])
            Sequence([1, 3])
            Sequence([2, 3])
            Sequence([1, 2, 3])
            Sequence([4])
            Sequence([1, 4])
            Sequence([2, 4])
            Sequence([1, 2, 4])
            Sequence([3, 4])
            Sequence([1, 3, 4])
            Sequence([2, 3, 4])
            Sequence([1, 2, 3, 4])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([1, 2, 3, 4])
            >>> for combination in enumerator.yield_combinations(
            ...     minimum_length=3,
            ...     ):
            ...     combination
            Sequence([1, 2, 3])
            Sequence([1, 2, 4])
            Sequence([1, 3, 4])
            Sequence([2, 3, 4])
            Sequence([1, 2, 3, 4])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([1, 2, 3, 4])
            >>> for combination in enumerator.yield_combinations(
            ...     maximum_length=2,
            ...     ):
            ...     combination
            Sequence([])
            Sequence([1])
            Sequence([2])
            Sequence([1, 2])
            Sequence([3])
            Sequence([1, 3])
            Sequence([2, 3])
            Sequence([4])
            Sequence([1, 4])
            Sequence([2, 4])
            Sequence([3, 4])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([1, 2, 3, 4])
            >>> for combination in enumerator.yield_combinations(
            ...     minimum_length=2,
            ...     maximum_length=2,
            ...     ):
            ...     combination
            Sequence([1, 2])
            Sequence([1, 3])
            Sequence([2, 3])
            Sequence([1, 4])
            Sequence([2, 4])
            Sequence([3, 4])

        ..  container:: example

            >>> enumerator = abjad.Enumerator('text')
            >>> for combination in enumerator.yield_combinations():
            ...     ''.join([str(_) for _ in combination])
            ''
            't'
            'e'
            'te'
            'x'
            'tx'
            'ex'
            'tex'
            't'
            'tt'
            'et'
            'tet'
            'xt'
            'txt'
            'ext'
            'text'

        Yields combinations in binary string order.

        Returns sequence generator.
        """
        import abjad
        length = len(self.sequence)
        for i in range(2**length):
            binary_string = abjad.mathtools.integer_to_binary_string(i)
            binary_string = binary_string.zfill(length)
            sublist = []
            for j, digit in enumerate(reversed(binary_string)):
                if digit == '1':
                    sublist.append(self.sequence[j])
            yield_sublist = True
            if minimum_length is not None:
                if len(sublist) < minimum_length:
                    yield_sublist = False
            if maximum_length is not None:
                if maximum_length < len(sublist):
                    yield_sublist = False
            if yield_sublist:
                if isinstance(self.sequence, str):
                    yield ''.join(sublist)
                else:
                    yield abjad.sequence(items=sublist)

    def yield_outer_product(self):
        """
        Yields outer product of sequences in sequence.

        ..  container:: example

            >>> sequences = [[1, 2, 3], ['a', 'b']]
            >>> enumerator = abjad.Enumerator(sequences)
            >>> for sequence_ in enumerator.yield_outer_product():
            ...     sequence_
            ...
            Sequence([1, 'a'])
            Sequence([1, 'b'])
            Sequence([2, 'a'])
            Sequence([2, 'b'])
            Sequence([3, 'a'])
            Sequence([3, 'b'])

        ..  container:: example

            >>> sequences = [[1, 2, 3], ['a', 'b'], ['X', 'Y']]
            >>> enumerator = abjad.Enumerator(sequences)
            >>> for sequence_ in enumerator.yield_outer_product():
            ...     sequence_
            ...
            Sequence([1, 'a', 'X'])
            Sequence([1, 'a', 'Y'])
            Sequence([1, 'b', 'X'])
            Sequence([1, 'b', 'Y'])
            Sequence([2, 'a', 'X'])
            Sequence([2, 'a', 'Y'])
            Sequence([2, 'b', 'X'])
            Sequence([2, 'b', 'Y'])
            Sequence([3, 'a', 'X'])
            Sequence([3, 'a', 'Y'])
            Sequence([3, 'b', 'X'])
            Sequence([3, 'b', 'Y'])

        ..  container:: example

            >>> sequences = [[1, 2, 3], [4, 5], [6, 7, 8]]
            >>> enumerator = abjad.Enumerator(sequences)
            >>> for sequence_ in enumerator.yield_outer_product():
            ...     sequence_
            ...
            Sequence([1, 4, 6])
            Sequence([1, 4, 7])
            Sequence([1, 4, 8])
            Sequence([1, 5, 6])
            Sequence([1, 5, 7])
            Sequence([1, 5, 8])
            Sequence([2, 4, 6])
            Sequence([2, 4, 7])
            Sequence([2, 4, 8])
            Sequence([2, 5, 6])
            Sequence([2, 5, 7])
            Sequence([2, 5, 8])
            Sequence([3, 4, 6])
            Sequence([3, 4, 7])
            Sequence([3, 4, 8])
            Sequence([3, 5, 6])
            Sequence([3, 5, 7])
            Sequence([3, 5, 8])

        Returns sequence generator.
        """
        def _helper(sequence_1, sequence_2):
            result = []
            for item_1 in sequence_1:
                for item_2 in sequence_2:
                    result.extend([item_1 + [item_2]])
            return result
        import abjad
        sequences = [abjad.sequence(_) for _ in self.sequence]
        sequences[0] = [[_] for _ in sequences[0]]
        result = functools.reduce(_helper, sequences)
        for element in result:
            yield abjad.sequence(items=element)

    def yield_pairs(self):
        """
        Yields pairs sequence items.

        ..  container:: example

            Without duplicate items:

            >>> enumerator = abjad.Enumerator([1, 2, 3, 4])
            >>> for pair in enumerator.yield_pairs():
            ...     pair
            ...
            Sequence([1, 2])
            Sequence([1, 3])
            Sequence([1, 4])
            Sequence([2, 3])
            Sequence([2, 4])
            Sequence([3, 4])

        ..  container:: example

            With duplicate items:

            >>> enumerator = abjad.Enumerator([1, 1, 1])
            >>> for pair in enumerator.yield_pairs():
            ...     pair
            ...
            Sequence([1, 1])
            Sequence([1, 1])
            Sequence([1, 1])

        ..  container:: example

            Length-1 sequence:

            >>> enumerator = abjad.Enumerator([1])
            >>> for pair in enumerator.yield_pairs():
            ...     pair
            ...

        ..  container:: example

            Empty sequence:

            >>> enumerator = abjad.Enumerator([])
            >>> for pair in enumerator.yield_pairs():
            ...     pair
            ...

        Returns generator of length-2 sequences.
        """
        import abjad
        for i, item in enumerate(self.sequence):
            start = i + 1
            for item_ in self.sequence[start:]:
                pair = [item, item_]
                yield abjad.sequence(items=pair)

    def yield_partitions(self):
        """
        Yields partitions of sequence.

        ..  container:: example

            >>> enumerator = abjad.Enumerator([0, 1, 2])
            >>> for partition in enumerator.yield_partitions():
            ...     partition
            ...
            Sequence([Sequence([0, 1, 2])])
            Sequence([Sequence([0, 1]), Sequence([2])])
            Sequence([Sequence([0]), Sequence([1, 2])])
            Sequence([Sequence([0]), Sequence([1]), Sequence([2])])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([0, 1, 2, 3])
            >>> for partition in enumerator.yield_partitions():
            ...     partition
            ...
            Sequence([Sequence([0, 1, 2, 3])])
            Sequence([Sequence([0, 1, 2]), Sequence([3])])
            Sequence([Sequence([0, 1]), Sequence([2, 3])])
            Sequence([Sequence([0, 1]), Sequence([2]), Sequence([3])])
            Sequence([Sequence([0]), Sequence([1, 2, 3])])
            Sequence([Sequence([0]), Sequence([1, 2]), Sequence([3])])
            Sequence([Sequence([0]), Sequence([1]), Sequence([2, 3])])
            Sequence([Sequence([0]), Sequence([1]), Sequence([2]), Sequence([3])])

        Returns generator of nested sequences.
        """
        import abjad
        length = len(self.sequence) - 1
        for i in range(2**length):
            binary_string = abjad.mathtools.integer_to_binary_string(i)
            binary_string = binary_string.zfill(length)
            part = list(self.sequence[0:1])
            partition = [part]
            for n, token in zip(self.sequence[1:], binary_string):
                if int(token) == 0:
                    part.append(n)
                else:
                    part = [n]
                    partition.append(part)
            parts = [abjad.sequence(items=_) for _ in partition]
            partition = abjad.sequence(items=parts)
            yield partition

    def yield_permutations(self):
        """
        Yields permutations of sequence.

        ..  container:: example

            >>> enumerator = abjad.Enumerator([1, 2, 3])
            >>> for permutation in enumerator.yield_permutations():
            ...     permutation
            ...
            Sequence([1, 2, 3])
            Sequence([1, 3, 2])
            Sequence([2, 1, 3])
            Sequence([2, 3, 1])
            Sequence([3, 1, 2])
            Sequence([3, 2, 1])

        Returns sequence generator.
        """
        import abjad
        length = len(self.sequence)
        for permutation in itertools.permutations(tuple(range(length))):
            permutation = abjad.sequence(items=permutation)
            yield self.sequence.permute(permutation)

    def yield_set_partitions(self):
        """
        Yields set partitions of sequence.

        ..  container:: example

            >>> enumerator = abjad.Enumerator([21, 22, 23, 24])
            >>> for set_partition in enumerator.yield_set_partitions():
            ...     set_partition
            ...
            Sequence([Sequence([21, 22, 23, 24])])
            Sequence([Sequence([21, 22, 23]), Sequence([24])])
            Sequence([Sequence([21, 22, 24]), Sequence([23])])
            Sequence([Sequence([21, 22]), Sequence([23, 24])])
            Sequence([Sequence([21, 22]), Sequence([23]), Sequence([24])])
            Sequence([Sequence([21, 23, 24]), Sequence([22])])
            Sequence([Sequence([21, 23]), Sequence([22, 24])])
            Sequence([Sequence([21, 23]), Sequence([22]), Sequence([24])])
            Sequence([Sequence([21, 24]), Sequence([22, 23])])
            Sequence([Sequence([21]), Sequence([22, 23, 24])])
            Sequence([Sequence([21]), Sequence([22, 23]), Sequence([24])])
            Sequence([Sequence([21, 24]), Sequence([22]), Sequence([23])])
            Sequence([Sequence([21]), Sequence([22, 24]), Sequence([23])])
            Sequence([Sequence([21]), Sequence([22]), Sequence([23, 24])])
            Sequence([Sequence([21]), Sequence([22]), Sequence([23]), Sequence([24])])

        Returns set partitions in order of restricted growth function.

        Returns generator of list of lists.
        """
        length = len(self.sequence)
        for rgf in Enumerator._yield_restricted_growth_functions(length):
            partition = Enumerator._partition_by_rgf(self.sequence, rgf)
            yield partition

    def yield_subsequences(self, minimum_length=0, maximum_length=None):
        """
        Yields subsequences of ``sequence``.

        ..  container:: example

            >>> enumerator = abjad.Enumerator([0, 1, 2])
            >>> for subsequence in enumerator.yield_subsequences():
            ...     subsequence
            ...
            Sequence([])
            Sequence([0])
            Sequence([0, 1])
            Sequence([0, 1, 2])
            Sequence([1])
            Sequence([1, 2])
            Sequence([2])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([0, 1, 2, 3, 4])
            >>> for subsequence in enumerator.yield_subsequences(
            ...     minimum_length=3,
            ...     ):
            ...     subsequence
            ...
            Sequence([0, 1, 2])
            Sequence([0, 1, 2, 3])
            Sequence([0, 1, 2, 3, 4])
            Sequence([1, 2, 3])
            Sequence([1, 2, 3, 4])
            Sequence([2, 3, 4])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([0, 1, 2, 3, 4])
            >>> for subsequence in enumerator.yield_subsequences(
            ...     maximum_length=3,
            ...     ):
            ...     subsequence
            ...
            Sequence([])
            Sequence([0])
            Sequence([0, 1])
            Sequence([0, 1, 2])
            Sequence([1])
            Sequence([1, 2])
            Sequence([1, 2, 3])
            Sequence([2])
            Sequence([2, 3])
            Sequence([2, 3, 4])
            Sequence([3])
            Sequence([3, 4])
            Sequence([4])

        ..  container:: example

            >>> enumerator = abjad.Enumerator([0, 1, 2, 3, 4])
            >>> for subsequence in enumerator.yield_subsequences(
            ...     minimum_length=3,
            ...     maximum_length=3,
            ...     ):
            ...     subsequence
            ...
            Sequence([0, 1, 2])
            Sequence([1, 2, 3])
            Sequence([2, 3, 4])

        Returns sequence generator.
        """
        length = len(self.sequence)
        if maximum_length is None:
            maximum_length = length
        for i in range(length):
            start_j = minimum_length + i
            stop_j = min(maximum_length + i, length) + 1
            for j in range(start_j, stop_j):
                if i < j or i == 0:
                    subsequence = self.sequence[i:j]
                    yield subsequence



import typing
from abjad.system.AbjadValueObject import AbjadValueObject


class BowMotionTechnique(AbjadValueObject):
    """
    Bow motion technique.

    ..  container:: example

        Jété:

        >>> bow_motion_technique = abjad.BowMotionTechnique('jete')
        >>> abjad.f(bow_motion_technique)
        abjad.BowMotionTechnique(
            technique_name='jete',
            )

    ..  container:: example

        Ordinario:

        >>> bow_motion_technique = abjad.BowMotionTechnique('ordinario')
        >>> abjad.f(bow_motion_technique)
        abjad.BowMotionTechnique(
            technique_name='ordinario',
            )

    Valid technique names include 'ordinario', 'jeté' and 'circular'.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_technique_name',
        )

    _parameter = True

    _publish_storage_format = True

    _valid_technique_names = (
        'circular',
        'jete',
        'ordinario',
        None,
        )

    ### INITIALIZER ###

    def __init__(
        self,
        technique_name: str = None,
        ) -> None:
        assert technique_name in self._valid_technique_names
        self._technique_name = technique_name

    ### PUBLIC PROPERTIES ###

    @property
    def glissando_style(self) -> str:
        """
        Gets glissando style of bow motion technique.

        ..  container:: example

            >>> abjad.BowMotionTechnique('jete').glissando_style
            'dotted-line'

        """
        if self.technique_name == 'circular':
            return 'zigzag'
        elif self.technique_name == 'jete':
            return 'dotted-line'
        return 'line'

    @property
    def parameter(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.BowMotionTechnique('jete').parameter
            True

        """
        return self._parameter

    @property
    def technique_name(self) -> typing.Optional[str]:
        """
        Gets technique name of bow motion technique.

        ..  container:: example

            >>> abjad.BowMotionTechnique('jete').technique_name
            'jete'

        """
        return self._technique_name

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on bow motion technique.
        """
        pass



import functools
import numbers
from . import constants
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.FormatSpecification import FormatSpecification


@functools.total_ordering
class StaffPosition(AbjadValueObject):
    """
    Staff position.

    ..  container:: example

        Initializes staff position at middle line of staff:

        >>> abjad.StaffPosition(0)
        StaffPosition(0)

    ..  container:: example

        Initializes staff position one space below middle line of staff:

        >>> abjad.StaffPosition(-1)
        StaffPosition(-1)

    ..  container:: example

        Initializes staff position one line below middle line of staff:

        >>> abjad.StaffPosition(-2)
        StaffPosition(-2)

    ..  container:: example

        Initializes from other staff position:

        >>> staff_position = abjad.StaffPosition(-2)
        >>> abjad.StaffPosition(staff_position)
        StaffPosition(-2)

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_number',
        )

    ### INITIALIZER ###

    def __init__(self, number=0):
        if isinstance(number, type(self)):
            number = number.number
        assert isinstance(number, numbers.Number), repr(number)
        self._number = number

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when `argument` is a staff position with the same number as
        this staff position.

        ..  container:: example

            >>> staff_position_1 = abjad.StaffPosition(-2)
            >>> staff_position_2 = abjad.StaffPosition(-2)
            >>> staff_position_3 = abjad.StaffPosition(0)

            >>> staff_position_1 == staff_position_1
            True
            >>> staff_position_1 == staff_position_2
            True
            >>> staff_position_1 == staff_position_3
            False

            >>> staff_position_2 == staff_position_1
            True
            >>> staff_position_2 == staff_position_2
            True
            >>> staff_position_2 == staff_position_3
            False

            >>> staff_position_3 == staff_position_1
            False
            >>> staff_position_3 == staff_position_2
            False
            >>> staff_position_3 == staff_position_3
            True

        Returns true or false.
        """
        return super().__eq__(argument)

    def __hash__(self):
        """
        Hashes staff position.

        Returns integer.
        """
        return super().__hash__()

    def __lt__(self, argument):
        """
        Is true when staff position is less than `argument`.

        ..  container:: example

            >>> staff_position_1 = abjad.StaffPosition(-2)
            >>> staff_position_2 = abjad.StaffPosition(-2)
            >>> staff_position_3 = abjad.StaffPosition(0)

            >>> staff_position_1 < staff_position_1
            False
            >>> staff_position_1 < staff_position_2
            False
            >>> staff_position_1 < staff_position_3
            True

            >>> staff_position_2 < staff_position_1
            False
            >>> staff_position_2 < staff_position_2
            False
            >>> staff_position_2 < staff_position_3
            True

            >>> staff_position_3 < staff_position_1
            False
            >>> staff_position_3 < staff_position_2
            False
            >>> staff_position_3 < staff_position_3
            False

        Returns true or false.
        """
        try:
            argument = type(self)(argument)
        except Exception:
            return False
        return self.number < argument.number

    def __str__(self):
        """
        Gets string representation of staff position.

        ..  container:: example

            >>> str(abjad.StaffPosition(-2))
            'StaffPosition(-2)'

        Returns string.
        """
        return '{}({})'.format(type(self).__name__, self.number)

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_is_indented=False,
            storage_format_args_values=[self.number],
            storage_format_kwargs_names=[],
            )

    ### PUBLIC PROPERTIES ###

    @property
    def number(self):
        """
        Gets staff position number.

        ..  container:: example

            >>> abjad.StaffPosition(-2).number
            -2

        Returns number.
        """
        return self._number

    ### PUBLIC METHODS ###

    def to_pitch(self, clef='treble'):
        """
        Makes named pitch from staff position and `clef`.

        ..  container:: example

            Treble clef:

            >>> clef = abjad.Clef('treble')
            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch(clef=clef)
            ...     message = '{!s}\t{}'.format(staff_position, pitch)
            ...     print(message)
            ...
            StaffPosition(-6)	c'
            StaffPosition(-5)	d'
            StaffPosition(-4)	e'
            StaffPosition(-3)	f'
            StaffPosition(-2)	g'
            StaffPosition(-1)	a'
            StaffPosition(0)	b'
            StaffPosition(1)	c''
            StaffPosition(2)	d''
            StaffPosition(3)	e''
            StaffPosition(4)	f''
            StaffPosition(5)	g''

        ..  container:: example

            Bass clef:

            >>> clef = abjad.Clef('bass')
            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch(clef=clef)
            ...     message = '{!s}\t{}'.format(staff_position, pitch)
            ...     print(message)
            ...
            StaffPosition(-6)	e,
            StaffPosition(-5)	f,
            StaffPosition(-4)	g,
            StaffPosition(-3)	a,
            StaffPosition(-2)	b,
            StaffPosition(-1)	c
            StaffPosition(0)	d
            StaffPosition(1)	e
            StaffPosition(2)	f
            StaffPosition(3)	g
            StaffPosition(4)	a
            StaffPosition(5)	b

        ..  container:: example

            Alto clef:

            >>> clef = abjad.Clef('alto')
            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch(clef=clef)
            ...     message = '{!s}\t{}'.format(staff_position, pitch)
            ...     print(message)
            ...
            StaffPosition(-6)	d
            StaffPosition(-5)	e
            StaffPosition(-4)	f
            StaffPosition(-3)	g
            StaffPosition(-2)	a
            StaffPosition(-1)	b
            StaffPosition(0)	c'
            StaffPosition(1)	d'
            StaffPosition(2)	e'
            StaffPosition(3)	f'
            StaffPosition(4)	g'
            StaffPosition(5)	a'

        ..  container:: example

            Percussion clef:

            >>> clef = abjad.Clef('percussion')
            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch(clef=clef)
            ...     message = '{!s}\t{}'.format(staff_position, pitch)
            ...     print(message)
            ...
            StaffPosition(-6)	d
            StaffPosition(-5)	e
            StaffPosition(-4)	f
            StaffPosition(-3)	g
            StaffPosition(-2)	a
            StaffPosition(-1)	b
            StaffPosition(0)	c'
            StaffPosition(1)	d'
            StaffPosition(2)	e'
            StaffPosition(3)	f'
            StaffPosition(4)	g'
            StaffPosition(5)	a'

        Returns new named pitch.
        """
        import abjad
        clef = abjad.Clef(clef)
        offset_staff_position_number = self.number
        offset_staff_position_number -= clef.middle_c_position.number
        offset_staff_position = type(self)(offset_staff_position_number)
        octave_number = offset_staff_position.number // 7 + 4
        diatonic_pc_number = offset_staff_position.number % 7
        pitch_class_number = constants._diatonic_pc_number_to_pitch_class_number[
            diatonic_pc_number]
        pitch_number = 12 * (octave_number - 4)
        pitch_number += pitch_class_number
        named_pitch = abjad.NamedPitch(pitch_number)
        return named_pitch
